/*
 * Copyright (C) 2021-2022 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#ifndef __COMMON_H__
#define __COMMON_H__

#if defined(__QNX__)
#   include <sys/syspage.h> // SYSPAGE_ENTRY
#endif

#if defined (_WIN32) || defined (_WIN64)
#   include <windows.h>
/* Some MinGW/CYGWIN distributions don't define the macro: */
#   ifndef ENABLE_VIRTUAL_TERMINAL_PROCESSING
#       define ENABLE_VIRTUAL_TERMINAL_PROCESSING 0x0004
#   endif
#endif

#include <sys/time.h>   // gettimeofday()
#include <time.h>       // localtime()

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>     // memset()
#include <stdint.h>     // uint8_t, ...
#include <inttypes.h>   // PRIx64
#include <stdbool.h>    // bool
#include <pthread.h>    // pthread()
#include <unistd.h>     // usleep()
#include <stdarg.h>     // args

#include "libpowerfip.h"

#define APP_VERSION "1.0.0"

#pragma pack(1) // 1 byte alignment

struct tst_measure {
    /* synchro callback measure */
    uint64_t nb_sync_cb;
    uint64_t cumul_sync_cb_duration;
    uint64_t min_sync_cb_duration;
    uint64_t max_sync_cb_duration;
    uint64_t min_sync_cb_last;
    uint64_t max_sync_cb_last;
    /* fip variables stats */
    uint64_t r_var_ok;
    uint64_t w_var_ok;
    uint64_t r_var_warn;
    uint64_t w_var_warn;
    /*   -> ping/pong test */
    uint64_t pp_ok;
    uint64_t pp_nok;
    uint64_t min_pp_diff;
    uint64_t max_pp_diff;
};

#define TST_MAX_VAR_COUNT 3
#define TST_MAX_BA_MCYCLE_COUNT 1
struct tst_node {
    struct pwrfip_node *n;
    /* aele configuration */
    struct pwrfip_aele *al;
    /* consumed variables list */
    struct pwrfip_var *cons_varlist[TST_MAX_VAR_COUNT];
    int cons_varcnt;
    /* produced variables list */
    struct pwrfip_var *prod_varlist[TST_MAX_VAR_COUNT];
    int prod_varcnt;
    /* synchro variables list */
    struct pwrfip_var *sync_varlist[TST_MAX_VAR_COUNT];
    int sync_varcnt;
    /* bus arbiter: macrocycles list */
    struct pwrfip_ba_mcycle *mc_list[TST_MAX_BA_MCYCLE_COUNT];
    int mc_cnt;
};

#define PING_STATION 0
#define PONG_STATION 1

struct tst_session {
    /* pci device handle */
    struct pwrfip_dev *dev;
    uint8_t station_number;
    uint8_t station_bitrate;
    uint32_t station_tr_ustime;
    uint32_t station_ts_ustime;
    uint32_t ba_cycle_ustime;
    uint32_t ba_midcycle_ustime;
    /* fip node instance (handles) */
    struct tst_node node;
    /* performance measurement */
    struct tst_measure stats;
};
#pragma pack()

extern int running;
extern bool dbg_coloriz;

/****************************************************************************************
 * FIP STATION FUNCTIONS
 ***************************************************************************************/
extern int tst_session_start(struct tst_session *sess);
extern int tst_session_stop(struct tst_session *sess);
extern int tst_node_report(struct tst_session *sess);

/****************************************************************************************
 * FIP BOARDS FUNCTIONS
 ***************************************************************************************/
extern int dev_open(int dev_idx, struct tst_session *sess);
extern int dev_close(struct tst_session *sess);
extern void dev_show_all();

/****************************************************************************************
 * STATISTICS
 ***************************************************************************************/
extern void update_min_max(uint64_t current_value, uint64_t *min_value,
            uint64_t *max_value);
extern int _usleep(uint32_t usec);
#if defined(__QNX__) || defined(_WIN32) || defined(_WIN64)
extern int _gettime(uint64_t *when);
extern void update_stats(uint64_t *start, uint64_t *stop, uint64_t *cp,
    uint64_t *cumul, uint64_t *min, uint64_t *max);
#else
extern int _gettime(struct timeval *when);
extern void update_stats(struct timeval *start, struct timeval *stop,
    uint64_t *cp, uint64_t *cumul, uint64_t *min, uint64_t *max);
#endif

/****************************************************************************************
 * TIME/DATE FUNCTIONS
 ***************************************************************************************/
#define DBG_PRINT_INFO 1
#define DBG_PRINT_WARN 1
#define DBG_PRINT_ERR  1
#define DBG_PRINT_EVT  1
#define DBG_DUMP_MEM   0
#define DGB_COLORIZ    1

typedef enum
{
    sta0 = 0,   // station0
    sta1,       // station1
    app,        // application (general)
    no_src      // unknown
} msg_src;

typedef enum
{
    info = 0,   // information
    warn,       // warning
    err,        // error
    evt,        // event (callback)
    no_type     // unknown
} msg_type;

extern char *get_date();
extern struct timespec timespec_diff(struct timespec start, struct timespec end);
extern struct timespec timespec_add_ns(struct timespec ref, long ns);
extern bool b_then_a(struct timespec a, struct timespec b);

#if defined(__QNX__)
extern uint64_t clock_diff_ns(uint64_t start_clk_cycl, uint64_t end_clk_cycl);
#endif

#if defined(__QNX__) || defined(_WIN32) || defined(_WIN64)
extern unsigned long get_elapsed_us(uint64_t end, uint64_t start);
#else
extern unsigned long get_elapsed_us(struct timeval end, struct timeval start);
#endif

/****************************************************************************************
 * CUSTOMIZED PRINT FUNCTION
 ***************************************************************************************/
extern int _print(msg_src src, msg_type type, const char *fmt, ...);

#if defined (_WIN32) || defined (_WIN64)
extern void enable_console_colorization(); /* enable console colorization for win */
#endif

#endif