/*
 * Copyright (C) 2022 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 *
 * Description:
 * UA pong example (loopback subcriber value to publisher)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>     // memset()
#include <stdint.h>     // uint8_t, ...
#include <fcntl.h>      // open()
#include <unistd.h>     // read()
#include <libgen.h>     // basename
#include <inttypes.h>   // PRIx64+
#include <signal.h>     // signal
#include <time.h>       // nanosleep, timespec
#if defined (_WIN32) || defined (_WIN64)
#include <conio.h>      // kbhit()
#endif

#include "test.h"

char *app_name = NULL;
int running = 0;
struct tst_session sess;

static void help();
void handle_signal(int sig);
extern void dev_show_all();

int main(int argc, char *argv[])
{
    int ret = 0;
    int value = 0;
    struct timespec rqtp = {0};
    struct timespec rmtp = {0};

#if defined (_WIN32) || defined (_WIN64)
    SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
    enable_console_colorization();
#else
    uint8_t buff = 0;
#endif

    app_name = basename(argv[0]);

    /* Try to process all command line arguments */
    while ((value = getopt(argc, argv, "hv")) != -1) {
        switch (value) {
            case 'h':
                help();
                return EXIT_SUCCESS;
            case 'v':
                printf("%s %s\n", app_name, APP_VERSION);
                return EXIT_SUCCESS;
            case '?':
            default:
                help();
                return EXIT_FAILURE;
        }
    }

    signal(SIGINT, handle_signal);

    memset(&sess, 0, sizeof(struct tst_session));

    /* init settings here */
    opc_init(&sess);

    /* opcua starting ... */
    if (opc_pub_start(&sess) < 0)
        goto end;

#if defined(__linux__) || defined(__QNX__)
    /* set stdin non-blocking */
    fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
#endif

    rqtp.tv_sec = 0;
    rqtp.tv_nsec = 500000000; /* 500ms */

    /* main loop */
    /* TODO: wait a signal */
    _usleep(2000); //--!!
    _print(app, info,
        "[ok] app is active. press enter to close...\n");

    running = 1;
    while (running) {

        /* wait for a pressed key */
#if defined (_WIN32) || defined (_WIN64)
        if (kbhit())
#else
        if (read(STDIN_FILENO, &buff, 1) == 1)
#endif
        {
            running = 0;
        }

        ret = nanosleep(&rqtp, &rmtp);
    }
    printf("\n");
    ret = 0;

end:
    ret |= opc_pub_stop(&sess);
    if (ret)
        _print(opc, err, "[failed] opcua publisher stop\n");
    else
        _print(opc, info, "[ok] opcua publisher stop\n");

    if (!ret)
        _print(app, info, "[ok] app is stopped\n");

    return ret;
}

/**
 * \brief Callback function for handling signals.
 * \param sig identifier of signal
 */
void handle_signal(int sig)
{
    if (sig == SIGINT) {
        _print(app, info, "SIGINT signal received\n");
        running = 0;
        /* Reset signal handling to default behavior */
        signal(SIGINT, SIG_DFL);
    } else
        _print(app, warn, "signal received but not supported (sig=%d)\n", sig);
}

/**
 * \brief Print help for this application
 */
static void help()
{
    printf("\n");
    printf("Usage: %s [OPTION]...\n\n", app_name);
    printf("UA pong application.\n\n");
    printf("Options:\n");
    printf("    -h   show this help and exit\n");
    printf("    -v   show version and exit\n");
    printf("\n");
    printf("Examples: \n");
    printf("    %s\n", app_name);
    printf("\n");
}